unit FSpectrum;
(*==========================================================================
  WAV file spectrum viewer
    This is an example of the use of the FFT and FFFT routines.
    This form reads a WAV sound file and plots the data.
    It calculates and shows the freqeuency spectrum of a part of the signal.
    This example demonstrates:
      - The FFT routine works with any signal length
        not necessarily a power of 2
      - The FFFT (Fractional Fast Fourier Transform) routine can be
        used to get a fractinoal part (zoom) of the spectrum.
      - How weighting windows affect the power spectrum
--------------------------------------------------------------------------
  written by   : Andreas Hofer,  fft@aho.ch,  http://www.aho.ch
  version      : 1.0.1,  18.3.2002
  copyright    : Andreas Hofer Software (c) 2002
============================================================================*)

{$BOOLEVAL OFF }
{$EXTENDEDSYNTAX ON }
{$IOCHECKS ON }
{$LONGSTRINGS ON }
{$OPENSTRINGS ON }
{$TYPEDADDRESS OFF }
{$VARSTRINGCHECKS ON }
{$WRITEABLECONST OFF }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, MPlayer, ComCtrls, Buttons,

  WavFile, AHoFFT;


type
  TSpectrumForm = class(TForm)
    OpenDialog1: TOpenDialog;
    Panel1: TPanel;
    Button1: TButton;
    Panel3: TPanel;
    Splitter1: TSplitter;
    Splitter2: TSplitter;
    Panel5: TPanel;
    Label5: TLabel;
    PaintBox3: TPaintBox;
    Panel6: TPanel;
    Panel2: TPanel;
    PaintBox1: TPaintBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Splitter3: TSplitter;
    Panel4: TPanel;
    PaintBox2: TPaintBox;
    Label12: TLabel;
    Label6: TLabel;
    Panel7: TPanel;
    PaintBox4: TPaintBox;
    Label8: TLabel;
    Label7: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    CheckBox1: TCheckBox;
    ComboBox1: TComboBox;
    Label13: TLabel;
    MediaPlayer1: TMediaPlayer;
    Button2: TButton;
    Label16: TLabel;
    Label17: TLabel;
    Label18: TLabel;
    Label19: TLabel;
    Label14: TLabel;
    Label15: TLabel;
    procedure Button1Click(Sender: TObject);
    procedure PaintBox1Paint(Sender: TObject);
    procedure PaintBox2Paint(Sender: TObject);
    procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox3Paint(Sender: TObject);
    procedure PaintBox3MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBox3MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox3MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox4Paint(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ComboBox1Change(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    signal : array of Single;
    sampleRate : Integer;
    sigLeft : Integer;  // section border left
    sigRight : Integer; // section border right
    sigDrawn : Boolean; // flag for drawing the section lines
    sigDrag : Integer;  // dragging section 0:nothing, 1:left, 2:right, 3:both
    spectLeft : Integer;  // spectrum section border left
    spectRight : Integer; // spectrum section border right
    spectDrawn : Boolean; // flag for drawing the spectrum section lines
    spectDrag : Integer;  // 0:nothing, 1:left, 2:right, 3:both
    xDrag : Integer;    // x pos of start of drag
    drLeft, drRight : Integer; // original section borders while dragging
    procedure DrawSigSect;
    procedure SetSigSect(a,b : Integer);
    procedure DrawSpectSect;
    procedure SetSpectSect(a,b : Integer);
    procedure ApplyDataWindow(var signal : array of TFFTFloat);
  end;

var
  SpectrumForm: TSpectrumForm;

implementation

{$R *.DFM}


procedure PlotTrace(const trace : array of Single;
                    cv : TCanvas;
                    rect : TRect;
                    yMin,yMax : Single;
                    logY : Boolean);
  var
    i,x,xp,y,yl,yh,w,h,np : Integer;
    points : array of TPoint;
    c,lny : Single;

  procedure Add;
  begin
    if xp >= 0 then begin
      points[np] := Point(xp+rect.left, rect.bottom-yl);
      Inc(np);
      if yl <> yh then begin
        points[np] := Point(xp+rect.left, rect.bottom-yh);
        Inc(np);
      end;
    end;
    yl := h+1;
    yh := -1;
  end;

begin
  w := rect.right - rect.left;
  h := rect.bottom - rect.top;
  SetLength(points,2*w);
  np := 0;
  if logY then begin
    lny := ln(yMin);
    c := h / (ln(yMax) - lny);
  end
  else begin
    lny := 0;
    c := h / (yMax - yMin);
  end;
  xp := -1;
  for i := 0 to High(trace) do begin
    x := i*w div Length(trace);
    if logY then begin
      if trace[i] <= 0 then
        y := 0
      else begin
        y := Round((ln(trace[i]) - lny) * c);
      end;
    end  
    else begin
      y := Round((trace[i] - yMin) * c);
    end;
    if y < 0 then y := 0;
    if y > h-1 then y := h-1;
    if xp <> x then Add;
    xp := x;
    if y < yl then yl := y;
    if y > yh then yh := y;
  end;
  Add;
  SetLength(points,np);
  cv.Polyline(points);
end;


//--- private methods ------------------------------------------------------


procedure TSpectrumForm.DrawSigSect;
  var
    xl,xr : Integer;
begin
  if Length(signal) > 0 then begin
    with PaintBox1 do begin
      xl := sigLeft * Width div Length(signal);
      xr := sigRight * Width div Length(signal);
      Canvas.Pen.Color := clBlue;
      Canvas.Pen.Mode := pmNotXor;
      Canvas.MoveTo(xl,0);
      Canvas.LineTo(xl,Height);
      if xl <> xr then begin
        Canvas.MoveTo(xr,0);
        Canvas.LineTo(xr,Height);
      end;
      Canvas.Pen.Mode := pmCopy;
      sigDrawn := not sigDrawn;
    end;
  end;
end;



procedure TSpectrumForm.DrawSpectSect;
  var
    xl,xr : Integer;
begin
  if Length(signal) > 0 then begin
    with PaintBox3 do begin
      xl := spectLeft * Width div (sampleRate div 2);
      xr := spectRight * Width div (sampleRate div 2);
      Canvas.Pen.Color := clBlue;
      Canvas.Pen.Mode := pmNotXor;
      Canvas.MoveTo(xl,0);
      Canvas.LineTo(xl,Height);
      if xl <> xr then begin
        Canvas.MoveTo(xr,0);
        Canvas.LineTo(xr,Height);
      end;
      Canvas.Pen.Mode := pmCopy;
      spectDrawn := not spectDrawn;
    end;
  end;
end;



procedure TSpectrumForm.SetSigSect(a,b : Integer);
  var
    swap : Integer;
begin
  if a < 0 then a := 0;
  if a >= Length(signal) then a := Length(signal);
  if b < 0 then b := 0;
  if b >= Length(signal) then b := Length(signal);
  if a > b then begin
    swap := a;
    a := b;
    b := swap;
  end;
  if (sigLeft <> a) or (sigRight <> b) then begin
    if sigDrawn then DrawSigSect;
    sigLeft := a;
    sigRight := b;
    DrawSigSect;
    PaintBox2.Invalidate;
    if (sigDrag = 0) or (b-a < 8192) then begin
      PaintBox3.Invalidate;
      PaintBox4.Invalidate;
    end;
    Label6.Caption := IntToStr(b-a) + ' Samples, '
             +FloatToStr(Round(1e6*(b-a)/sampleRate)/1e6) + ' s';
  end;
end;



procedure TSpectrumForm.SetSpectSect(a,b : Integer);
  var
    swap : Integer;
begin
  if a < 0 then a := 0;
  if a >= sampleRate div 2 then a := sampleRate div 2;
  if b < 0 then b := 0;
  if b >= sampleRate div 2 then b := sampleRate div 2;
  if a > b then begin
    swap := a;
    a := b;
    b := swap;
  end;
  if (spectLeft <> a) or (spectRight <> b) then begin
    if spectDrawn then DrawSpectSect;
    spectLeft := a;
    spectRight := b;
    DrawSpectSect;
    if (spectDrag = 0) or (sigRight - sigLeft < 8192) then PaintBox4.Invalidate;
    Label10.Caption := IntToStr(spectLeft) + ' Hz';
    Label11.Caption := IntToStr(spectRight) + ' Hz';
  end;
end;


procedure TSpectrumForm.ApplyDataWindow(var signal : array of TFFTFloat);
  var
    wi : TFFTWindowInfo;
begin
  case Combobox1.ItemIndex of
    1 : wi := ApplySumCosineWindow(signal, wCoeffHann);
    2 : wi := ApplySumCosineWindow(signal, wCoeffHamming);
    3 : wi := ApplySumCosineWindow(signal, wCoeffBlackman);
    4 : wi := ApplySumCosineWindow(signal, wCoeffNuttall);
    5 : wi := ApplySumCosineWindow(signal, wCoeffFlattop3);
    6 : wi := ApplySumCosineWindow(signal, wCoeffTaylor60);
    7 : wi := ApplySumCosineWindow(signal, wCoeffTaylor80);
    8 : wi := ApplyWindow(signal, WTriangle);
    9 : wi := ApplyWindow(signal, WWelch);
    else Exit;
  end;
  // Correct the amplitude effects of the window
  ScaleVector(signal, 1/wi.intw);
end;


//--- Object inspector methods ---------------------------------------------


procedure TSpectrumForm.Button1Click(Sender: TObject);
  var
    len : Integer;
    rec : TWAVFormatRec;
begin
  if OpenDialog1.Execute then begin
    len := WAVGetLength(OpenDialog1.filename);
    SetLength(signal,len);
    WAVReadSamples(OpenDialog1.filename, 0, signal);

    Label1.Caption := IntToStr(len) + ' Samples';
    rec := WAVGetFormat(OpenDialog1.filename);
    sampleRate := rec.sampleRate;
    Label2.Caption := IntToStr(rec.sampleRate) + ' Hz, '
             + IntToStr(rec.bitsPerSample) + ' Bits';
    Label3.Caption := FloatToStr(Round(1e6*len/rec.sampleRate)/1e6) + ' s';
    SetSigSect(len div 3, len div 2);
    Label9.Caption := IntToStr(sampleRate div 2) + ' Hz';
    SetSpectSect(0, sampleRate div 8);

    PaintBox1.Invalidate;
    PaintBox2.Invalidate;
    PaintBox3.Invalidate;
    PaintBox4.Invalidate;

    MediaPlayer1.Close;
    MediaPlayer1.Filename := OpenDialog1.filename;
    MediaPlayer1.Open;
  end;
  Button2.Enabled := true;
end;

procedure TSpectrumForm.PaintBox1Paint(Sender: TObject);
begin
  with PaintBox1 do begin
    Canvas.FillRect(Rect(0,0,Width-1,Height-1));
    Canvas.Pen.Color := clBlack;
    Canvas.Pen.Mode := pmCopy;
    PlotTrace(signal, Canvas,
              Rect(0, 0, Width-1, Height-1),
              -1, +1, false);
    sigDrawn := false;
    DrawSigSect;
  end;
end;

procedure TSpectrumForm.PaintBox2Paint(Sender: TObject);
  var
    sl, srmsl : Integer;
begin
  with PaintBox2 do begin
    Canvas.FillRect(Rect(0,0,Width-1,Height-1));
    if sigRight-sigLeft > 0 then begin
      // Delphi 7 produced an internal error here. It seems to work
      // if we assign sigLeft and sigRight-sigLeft to local variables
      // first. Thanks to Jean Knoll who discovered this and the workaround.
      sl := sigLeft;
      srmsl := sigRight-sigLeft;
      PlotTrace(Copy(signal,sl, srmsl), Canvas,Rect(0, 0, Width-1, Height-1),-1, +1, false);
    end;
  end;
end;

procedure TSpectrumForm.PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
  var
    xl,xr,d : Integer;
begin
  if Length(signal) = 0 then Exit;
  with PaintBox1 do begin
    if sigDrag <> 0 then begin
      d := (x-xDrag) * Length(signal) div Width;
      case sigDrag of
        1 : SetSigSect(drLeft + d, sigRight);
        2 : SetSigSect(sigLeft, drRight + d);
        3 : SetSigSect(drLeft + d, drRight + d);
      end;
    end
    else begin
      xl := sigLeft * Width div Length(signal);
      xr := sigRight * Width div Length(signal);
      if (system.Abs(x-xl) < 3) or (system.Abs(x-xr) < 3) then
        Cursor := crHSplit
      else if (xl < x) and (x < xr) then
        Cursor := crSizeWE
      else begin
        Cursor := crDefault
      end;
    end;
  end;
end;

procedure TSpectrumForm.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
  var
    xl,xr : Integer;
begin
  if Button = mbLeft then begin
    if Length(signal) = 0 then Exit;
    with PaintBox1 do begin
      xl := sigLeft * Width div Length(signal);
      xr := sigRight * Width div Length(signal);
      drLeft := sigLeft;
      drRight := sigRight;
      xDrag := x;
      if (system.Abs(x-xl) < 3) then
        sigDrag := 1
      else if (system.Abs(x-xr) < 3) then
        sigDrag := 2
      else if (xl < x) and (x < xr) then
        sigDrag := 3
      else begin
        sigDrag := 0
      end;
    end;
  end;
end;


procedure TSpectrumForm.PaintBox1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  sigDrag := 0;
  PaintBox3.Invalidate;
  PaintBox4.Invalidate;
end;


procedure TSpectrumForm.PaintBox3Paint(Sender: TObject);
  var
    sig : array of Single;
    spectrum : array of TFFTComplex;
    amplitude : array of Single;
    max : Single;
    scale : Double;
    n,i : Integer;
begin
  with PaintBox3 do begin
    // Clear paint area
    Canvas.FillRect(Rect(0,0,Width-1,Height-1));

    // Get the section of the signal
    n := sigRight-sigLeft;
    if n < 2 then Exit;
    SetLength(sig, n);
    for i := 0 to n-1 do sig[i] := signal[sigLeft+i];

    ApplyDataWindow(sig);

    // Get the spectrum
    SetLength(spectrum, n);
    FFT(sig, spectrum);

    // Get the amplitude information from the spectrum
    SetLength(amplitude, n);
    RectToAbs(spectrum, amplitude);

    // Extract the first half of the spectrum (the second
    // half is not interesting since we have a real signal)
    SetLength(amplitude, n div 2 + 1);
    ScaleVector(amplitude, 2/n);
    amplitude[0] := amplitude[0] / 2;
    amplitude[n div 2] := amplitude[n div 2] / 2;

    // Search for an appropriate scale factor for the plot
    max := 0;
    for i := 0 to Length(amplitude)-1 do begin
      if amplitude[i] > max then max := amplitude[i];
    end;
    if max < 0.01 then scale := 0.01
    else if max < 0.02 then scale := 0.02
    else if max < 0.05 then scale := 0.05
    else if max < 0.1 then scale := 0.1
    else if max < 0.2 then scale := 0.2
    else if max < 0.5 then scale := 0.5
    else scale := 1;

    // Plot the amplitude
    Canvas.Pen.Color := clBlack;
    Canvas.Pen.Mode := pmCopy;
    if CheckBox1.Checked then begin
      PlotTrace(amplitude, Canvas,
                Rect(0, 0, Width-1, Height-1),
               scale/1e5, scale, true);
      Label17.Caption := FloatToStr(scale/1e5);
      Label16.Caption := FloatToStr(scale);
    end
    else begin
      PlotTrace(amplitude, Canvas,
                Rect(0, 0, Width-1, Height-1),
                0, scale, false);
      Label17.Caption := '0';
      Label16.Caption := FloatToStr(scale);
    end;

    // redraw the spectrum zoom lines
    spectDrawn := false;
    DrawSpectSect;
  end;
end;


procedure TSpectrumForm.PaintBox4Paint(Sender: TObject);
  var
    sig : array of Single;
    spectrum : array of TFFTComplex;
    amplitude : array of Single;
    freq1, freq2, df, ts : Double;
    max : Single;
    scale : Double;
    n,i : Integer;
begin
  with PaintBox4 do begin
    // Clear paint area
    Canvas.FillRect(Rect(0,0,Width-1,Height-1));

    // get spectrum section borders
    freq1 := spectLeft;
    freq2 := spectRight;
    if freq1 = freq2 then Exit;

    SetLength(spectrum, PaintBox4.width); // one spectrum line for each pixel
    df := (freq2-freq1) / Length(spectrum); // frequency step in spectrum
    ts := 1/sampleRate;

    // Get the section of the signal
    n := sigRight-sigLeft;
    if n = 0 then Exit;
    SetLength(sig, n);
    for i := 0 to n-1 do sig[i] := signal[sigLeft+i];

    ApplyDataWindow(sig);

    // Get the spectrum
    FFFT(sig,  // signal section to be analyzed
         sigLeft*ts,  // the signal section starts at this time
         ts,          // the signal has this sample rate
         spectrum, // the spectrum goes here
         freq1,    // the spectrum shall start at freq1
         df);      // the spectrum lines shall be spaced by df

    // Get the amplitude information from the spectrum
    SetLength(amplitude, Length(spectrum));
    RectToAbs(spectrum, amplitude);
    ScaleVector(amplitude, 2/n);

    // Search for an appropriate scale factor for the plot
    max := 0;
    for i := 0 to Length(amplitude)-1 do begin
      if amplitude[i] > max then max := amplitude[i];
    end;
    if max < 0.01 then scale := 0.01
    else if max < 0.02 then scale := 0.02
    else if max < 0.05 then scale := 0.05
    else if max < 0.1 then scale := 0.1
    else if max < 0.2 then scale := 0.2
    else if max < 0.5 then scale := 0.5
    else scale := 1;

    // Plot the amplitude
    Canvas.Pen.Color := clBlack;
    Canvas.Pen.Mode := pmCopy;
    if CheckBox1.Checked then begin
      PlotTrace(amplitude, Canvas,
                Rect(0, 0, Width-1, Height-1),
                scale/1e5, scale, true);
      Label19.Caption := FloatToStr(scale/1e5);
      Label18.Caption := FloatToStr(scale);
    end
    else begin
      PlotTrace(amplitude, Canvas,
                Rect(0, 0, Width-1, Height-1),
                0, scale, false);
      Label19.Caption := '0';
      Label18.Caption := FloatToStr(scale);
    end;
  end;
end;

procedure TSpectrumForm.PaintBox3MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  var
    xl,xr,d : Integer;
begin
  if Length(signal) = 0 then Exit;
  with PaintBox3 do begin
    if spectDrag <> 0 then begin
      d := (x-xDrag) * (sampleRate div 2) div Width;
      case spectDrag of
        1 : SetSpectSect(drLeft + d, spectRight);
        2 : SetSpectSect(spectLeft, drRight + d);
        3 : SetSpectSect(drLeft + d, drRight + d);
      end;
    end
    else begin
      xl := spectLeft * Width div (sampleRate div 2);
      xr := spectRight * Width div (sampleRate div 2);
      if (system.Abs(x-xl) < 3) or (system.Abs(x-xr) < 3) then
        Cursor := crHSplit
      else if (xl < x) and (x < xr) then
        Cursor := crSizeWE
      else begin
        Cursor := crDefault
      end;
    end;
  end;
end;

procedure TSpectrumForm.PaintBox3MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
  var
    xl,xr : Integer;
begin
  if Button = mbLeft then begin
    if Length(signal) = 0 then Exit;
    with PaintBox3 do begin
      xl := spectLeft * Width div (sampleRate div 2);
      xr := spectRight * Width div (sampleRate div 2);
      drLeft := spectLeft;
      drRight := spectRight;
      xDrag := x;
      if (system.Abs(x-xl) < 3) then
        spectDrag := 1
      else if (system.Abs(x-xr) < 3) then
        spectDrag := 2
      else if (xl < x) and (x < xr) then
        spectDrag := 3
      else begin
        spectDrag := 0
      end;
    end;
  end;
end;

procedure TSpectrumForm.PaintBox3MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  spectDrag := 0;
  PaintBox4.Invalidate;
end;


procedure TSpectrumForm.CheckBox1Click(Sender: TObject);
begin
  PaintBox3.Invalidate;
  PaintBox4.Invalidate;
end;

procedure TSpectrumForm.FormCreate(Sender: TObject);
begin
  ComboBox1.ItemIndex := 0;
  OpenDialog1.InitialDir := ExtractFilePath(ParamStr(0));
end;

procedure TSpectrumForm.ComboBox1Change(Sender: TObject);
begin
  PaintBox3.Invalidate;
  PaintBox4.Invalidate;
end;

procedure TSpectrumForm.Button2Click(Sender: TObject);
begin
  MediaPlayer1.Play;
end;

end.
